# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import csv
import os
import traceback
DEBUG = 1
TRY_TIMES = 30


class Reader(object):
    def __init__(self, col, colupper=None, delimiter=b"\t", quotechar=b"\"", valtype=str):
        self.col  = col
        if colupper is None or col > colupper:
            colupper = col
        self.delimiter = delimiter
        self.quotechar = quotechar
        self.colupper = colupper + 1
        self.valtype = valtype
    def setfile(self, fileobj):
        self.parser = csv.reader(fileobj, delimiter=self.delimiter, quotechar=self.quotechar)
        return self.parser
    def __iter__(self):
        return self
    def get_value(self, rows):
        return self.valtype(self.delimiter.join(rows[self.col:self.colupper]))
    def next(self):
        row = self.parser.next()
        val = self.get_value(row)
        return val, row

class Comparator(object):
    def __init__(self, lower=None, upper=None):
        self.lower = lower
        self.upper = upper
    def compare(self, val):
        lower = self.lower
        upper = self.upper
        if lower <= val <= upper:
            return 0
        elif val > upper:
            return -1
        return 1
    def __repr__(self):
        return "Comparator({}, {})".format(repr(self.lower), repr(self.upper))

class BaseBinSearch(object):
    def __init__(self, filepath, comparator, reader):
        self.filepath   = filepath
        self.comparator = comparator
        self.reader     = reader
        self.fio        = None
        self.buffer     = []
    def __enter__(self):
        filesize = os.path.getsize(self.filepath)
        self.fio = open(self.filepath)
        return self._binsearch(filesize, self.fio, self.comparator, self.reader)
    def __exit__(self, exc_type, exc_value, traceback):
        if self.fio:
            self.fio.close()
            self.fio = None

        if exc_value is None:
            return True
    def try_compare(self, fio, reader, comparator):
        parser = reader.setfile(fio)
        for i in range(TRY_TIMES):
            try:
                row = parser.next()
                val = reader.get_value(row)
                res = comparator.compare(val)
                if res==0:
                    self.buffer.append(row)
                return res
            except StopIteration:
                return None
            except:
                if DEBUG:
                    print(traceback.format_exc())

        return None
    def _iter_results(self, input_):
        if not self.buffer:
            raise StopIteration()

        for row in self.buffer:
            yield row

        self.reader.setfile(input_)
        comparator = self.comparator
        for val, row in self.reader:
            res = comparator.compare(val)
            if res == 0:
                yield row
            else:
                break

    def _binsearch(self, filesize, fio, comparator, reader):
        raise NotImplementedError()

class TextBinSearch(BaseBinSearch):
    def _binsearch(self, filesize, fio, comparator, reader):
        half = filesize // 2
        start = 0
        size = half
        while size > 100:
            fio.seek(half)
            fio.next()

            result = self.try_compare(fio, reader, comparator)
            if result is None:
                return None
            if result==0: #in
                break
            elif result < 0:
                pass
            else:
                start = half

            size = size // 2
            half = start + size

        return self._iter_results(fio)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
from . import bz2search
from .binsearch import BaseBinSearch

class BZ2BinSearch(BaseBinSearch):
    MINSIZE = 5000
    def __init__(self, filepath, comparator, reader):
        super(BZ2BinSearch, self).__init__(filepath, comparator, reader)
        self.prevblock = None
        self.fio = None
        self.blreader = None
    def _getblockreader(self, fio, binpos):
        if binpos < 0:
            return None

        if self.blreader is None:
            blreader = bz2search.BlockReader(fio, binpos)
            self.blreader = blreader
        else:
            self.blreader.reset(fio, binpos)
            blreader = self.blreader
        #skip 1 line
        try:
            blreader.next()
        except StopIteration:
            return None
        return blreader
    def _binsearch(self, filesize, fio, comparator, reader):
        if filesize <= self.MINSIZE:
            minsize = self.MINSIZE // filesize
        else:
            minsize = self.MINSIZE
        result = None
        half = filesize // 2
        start = 0
        size = half
        blreader = None
        # search block
        while size > minsize:
            binpos = bz2search.searchblock(fio, half)
            if(binpos<0) or \
                    (self.prevblock and self.prevblock == binpos):
                ## too large
                pass
            else:
                blreader = self._getblockreader(fio, binpos)
                if blreader:
                    result = self.try_compare(blreader, reader, comparator)
                    if result is None: #error
                        return None
                    if result == 0: #found
                        #result.reset()
                        break
                    elif result < 0:
                        #too large
                        pass
                    else:
                        #too small
                        start = half

            #print("b pos", binpos, self.prevblock, result, start, half, size)
            self.prevblock = binpos
            size = size // 2
            half = start + size

        #print("res", start, half, size)
        if size <= minsize and result < 0:
            binpos = bz2search.searchblock(fio, 0)
            blreader = self._getblockreader(fio, binpos)

        if blreader is None:
            return None

        if result==0:
            return self._iter_results(blreader)

        return self._binsearchinblock(blreader)
    def _binsearchinblock(self, blreader):
        while True:
            result = self.try_compare(blreader, self.reader, self.comparator)
            if result is None:
                return None
            if result == 0:
                return self._iter_results(blreader)
            # too large
            elif result < 0:
                return self._iter_results(blreader)
            else:
                continue





# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from setuptools import setup, find_packages, Extension
import os
from setuptools.command.test import test as TestCommand

class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True
    def run_tests(self):
        import pytest
        pytest.main(self.test_args)

shortdesc = "search (bz2)log file with binary search"

with open(os.path.join(os.path.dirname(__file__), "README.md")) as fd:
    longdesc = fd.read()
with open(os.path.join(os.path.dirname(__file__), "VERSION")) as fd:
    version = fd.read()
setup(name='logbinsearch',
      description=shortdesc,
      long_description=longdesc,
      version=version,
      author='takada-at',
      author_email='takada-at@klab.com',
      packages=find_packages(),
      ext_modules=[
          Extension('logbinsearch.bz2search', 
                    ['src/bz2search.c', 'src/microbunzip.c'], include_dirs=['src']),
      ],
      tests_require=['pytest'],
      cmdclass = {'test': PyTest},
      url = "https://github.com/takada-at/logbinsearch",
      scripts=['scripts/logdatetimesearch'],
)




# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import datetime
import os
import random
import bz2
from ..scripts import logdatetimesearch
from .. import binsearch

sample = os.path.join(os.path.dirname(__file__), 'sample/sample_datetime.bz2')

def generate(fromdate):
    if os.path.exists(sample):
        return sample

    bzf = bz2.BZ2File(sample, 'w', compresslevel=1)
    ptr = fromdate
    for i in range(200):
        for j in range(30):
            rows = [i*10 + j, ptr.strftime('%Y-%m-%d %H:%M:%S')] \
                + [random.randint(0, 100)]*10
            line = " ".join(map(bytes, rows)) + "\n"
            bzf.write(bytes(line))

        ptr += datetime.timedelta(minutes=2, seconds=10)

    return sample

def test_search():
    binsearch.DEBUG = 1
    results = []
    def mock(*string):
        #print(*string)
        results.append(string)

    fromdate = datetime.datetime(2015, 1, 20, 13, 0, 0)
    sample = generate(fromdate)
    logdatetimesearch.info = mock
    import sys
    def testsearch(fromd, tod):
        fromds = fromd.strftime('%Y-%m-%d %H:%M:%S')
        tods   = tod.strftime('%Y-%m-%d %H:%M:%S')
        sys.argv = ['logdatetimesearch.py', '-f', fromds, '-t', tods, sample]
        results = []
        logdatetimesearch.main()
        for row in results[1:10]:
            print(row)
            vals = row[0].split()
            vals = " ".join(vals[1:3])
            assert fromds <= vals
            assert tods >= vals

    fromd0 = datetime.datetime(2015, 1, 20, 16, 0, 0)
    tod0   = datetime.datetime(2015, 1, 20, 18, 16, 0)
    testsearch(fromd0, tod0)

    reload(logdatetimesearch)
    reload(sys)

